@@ -136,7 +136,11 @@ def evaluate_report(report, question=None, qitem=None, passall=False, verbose=Fa
             continue
         suite = loader.loadTestsFromTestCase(q)
         qtitle = q.question_title() if hasattr(q, 'question_title') else q.__qualname__
-        q_title_print = "Question %i: %s"%(n+1, qtitle)
+        if not report.abbreviate_questions:
+            q_title_print = "Question %i: %s"%(n+1, qtitle)
+        else:
+            q_title_print = "q%i) %s" % (n + 1, qtitle)
+
         print(q_title_print, end="")
         q.possible = 0
         q.obtained = 0
@@ -148,10 +152,20 @@ def evaluate_report(report, question=None, qitem=None, passall=False, verbose=Fa
         UTextResult.unmute = unmute # Hacky as well.
         UTextResult.setUpClass_time = q._cache.get(((q.__name__, 'setUpClass'), 'time'), 3) if hasattr(q, '_cache') and q._cache is not None else 3
 
+
         res = UTextTestRunner(verbosity=2, resultclass=UTextResult).run(suite)
         details = {}
         for s, msg in res.successes + res.failures + res.errors:
-            key = (q.__name__, s._testMethodName) # cannot use the cache_id method bc. it is not compatible with plain unittest.
+            # from unittest.suite import _ErrorHolder
+            # from unittest import _Err
+            # if isinstance(s, _ErrorHolder)
+            if hasattr(s, '_testMethodName'):
+                key = (q.__name__, s._testMethodName)
+            else:
+                # In case s is an _ErrorHolder (unittest.suite)
+                key = (q.__name__, s.id())
+            # key = (q.__name__, s._testMethodName) # cannot use the cache_id method bc. it is not compatible with plain unittest.
+
             detail = {}
             if (s,msg) in res.successes:
                 detail['status'] = "pass"
@@ -174,7 +188,7 @@ def evaluate_report(report, question=None, qitem=None, passall=False, verbose=Fa
         possible = res.testsRun
         obtained = len(res.successes)
 
-        assert len(res.successes) +  len(res.errors) + len(res.failures) == res.testsRun
+        # assert len(res.successes) +  len(res.errors) + len(res.failures) == res.testsRun
 
         obtained = int(w * obtained * 1.0 / possible ) if possible > 0 else 0
         score[n] = {'w': w, 'possible': w, 'obtained': obtained, 'items': details, 'title': qtitle, 'name': q.__name__,
